--[[============================================================================
 LAUFLICHT STEP SEQUENCER PLUGIN FOR RENOISE
 written and (c) by Thilo Geertzen alias Cie in 2010. http://www.cie-online.de

 CHANGE LOG
================================================================================
v0.41: 
  + fixed bug when starting stepsequencer gui again after closing
  + all settings are restored if stepsequencer is closed and reopened. (no real pattern reload yet)

v0.5:
  + removed local functions
  + instrument reloading/renaming/adding/deleting/doing-any-changes should update the 
  instrument selection in every track and should not result in an error message
  + Pattern-loading: loads the notes of the currently selected pattern into the step lights on gui startup. When changing the editstep, the lights of the step sequencer will be reloaded to the selected resolution on the fly.
  E.g. editstep 2 means: display of every second note; editstep 4 means: display of every fourth note
  Parameters like instrument number, note value, velocity, delay will be taken from the first occurence in the appropriate track,
  and the first three note columns are regarded only.
  
v0.6:
  + reload step sequencer lights on patternchange
  + paging: in editstep 1 and 2, implement pages to display the notes of the complete pattern. e.g. on editstep 1 page 1 shows pattern lines 1-15, page 2: lines 16-31 and so on. 
  + fixed http://www.renoise.com/board/index.php?showtopic=26552&view=findpost&p=209436
  + fixed http://www.renoise.com/board/index.php?showtopic=26552&view=findpost&p=210286

v0.61:
  + fixed http://www.renoise.com/board/index.php?showtopic=26552&view=findpost&p=211495
  
v0.63:
  + rewritten lights update in handle_app_idle_notification in  main.lua
  + fixed http://www.renoise.com/board/index.php?showtopic=26552&view=findpost&p=211498
  + fixed http://www.renoise.com/board/index.php?showtopic=26552&view=findpost&p=209421
  patternsize change can be done on the fly.
  + fixed some other minor bugs
v0.64
  + fixed http://www.renoise.com/board/index.php?showtopic=26552&view=findpost&p=211915
  + there is no limit on the number of tracks anymore. All audio tracks of the current song will be displayed.
  If a track is added/deleted during editing, the tool has to be reloaded to update the new amount of tracks,
  as there is no observer on these actions.

v1.0
  + rewritten gui drawing
  + it is possible to insert/delete a track on the fly. The gui will be updated (except insert a track). 
    when inserting a track, following Renoise lua Scripting API-bug appears: http://www.renoise.com/board/index.php?showtopic=27798
    workaround: reload pattern via switching editstep or close/restart gui
  + added delete button at the end of each track. This function deletes all notes of the pattern regarding the current editstep and selected page.
  + submitted to renoise app directory. :)
  
v.1.1
  + added Midi Mapping to steps. :) To map the steps of the Sequencer to a midi controller, click on Midi Map or Options->Midi Mapping
  and assign a your midi-controller to each step of each track. Please use controllers with the support of the function "switch" only, e.g. a toggleable button.
  + added MidiMapping to editstep and page. Please use controllers with support of absolute values only, e.g. knobs, sliders etc.
  The tool will automatically calculate the correct ranges of the appropriate values for the midi-controller.
  
v.1.11
  + fixed bug in midi mapping of steps. 
  + Now also all controller which send absolute values (knobs, sliders) can be used to switch on/off a step. A value <64 means off, >=64 on
  + fixed delay max value to ff

v.1.12
  + added Multisample support. Multisample Instruments will be shown in Instrument selection box like BankName:SampleName. If selected
  a sample from a bank, the basenote of the sample will be loaded in the Note1-selectbox by default.
  + patternloading: supports also multisamples. The first sample of the bank will be shown in the drop-down box if a multisample is detected in a track.
  + made Instr Box wider.

v.1.2
  + patternloading into sequencer: correct multisample loading: if basenote of multisample matches note in appropriate track, the correct 
  multisample will be shown in the drop-down box.
  + when changing an instrument in the drop-down box, all notes of the appropriate track in the pattern will be changed to the new selected instrument. 
    So you can instant change instruments without replacing the sample!
    
v.1.21
  + fixed http://www.renoise.com/board/index.php?showtopic=26552&view=findpost&p=219319
  A modal dialog box appears and shows an error message (added a new function that provides a modal dialog box with custom (error) message)
]]--

version = '1.21'

number_of_tracks_max = 8;
number_of_tracks=0;

number_of_steps = 16;
number_of_notes = 3;
editstep = 4;
editstepValue = 3;
editpage = 1;
lauflicht_start = 1; -- the startposition of the lauflicht. changes with editstep and page change
lauflicht_end = 64;
set_notes = true; -- set note only if clicked the button, not on load_pattern_into_sequencer()

vb = renoise.ViewBuilder() 


function load_tracks()

  number_of_tracks=0;
  for i=1,#renoise.song().tracks do 
    local track = renoise.song().tracks[i];
    if track.type==renoise.Track.TRACK_TYPE_SEQUENCER then
      number_of_tracks=number_of_tracks+1
    end
  end
  -- if a maximum number of displayable tracks is wanted, set number_of_tracks_max to the desired value and comment out the line below.
  if (number_of_tracks>number_of_tracks_max) then
    --number_of_tracks=number_of_tracks_max
  end

end


-- wlich gehts schneller wenn man renoise.song().pattern_iterator:lines_in_pattern(pattern_index) benutzt
-- dann keine iteration ber track mehr (tracknr schon im PatternTrackLine vorhanden)
function load_pattern_into_sequencer()
  
  local pattern_iter = renoise.song().pattern_iterator
  local pattern_index =  renoise.song().selected_pattern_index
  set_notes = false;
  
  
  for t=1, number_of_tracks do
    
   print ('Load Notes into Sequencer of Track ' ..t .. ' number of tracks=' ..number_of_tracks)
   local has_loaded_parameters = false;
   for pos,line in pattern_iter:lines_in_pattern_track(pattern_index,t) do
     -- print(line)
     local nc=1 -- check first notecolumn
     if not line.is_empty and line.note_columns[nc].note_value~=121 then -- do not parse lines with no note value
     
        if not line.note_columns[nc].is_empty then 
     
           -- YEAH Pattern loading!!!!!! :) :) :)
           -- alle lights positionen des track-document durchgehen und bei match den wert setzen           
           for lpos=1,#tracks[t].light_positions do
             --print ('track#' .. t .. ' position ' .. pos.line .. ' tracklight pos=' .. tracks[t].light_positions[lpos].value .. ' lpos='..lpos)
             --print ('note value ' .. line.note_columns[nc].note_value .. ' ' ..  line.note_columns[nc].instrument_value .. ' ' .. #renoise.song().instruments)
             if tracks[t].light_positions[lpos].value==pos.line then
               -- parameters: only load the first occurence in pattern
               if not has_loaded_parameters and line.note_columns[nc].note_value~=120 and 
                    line.note_columns[nc].instrument_value<#renoise.song().instruments then -- no note OFF and no empty instrument notes

                 -- loading instrument
                 --tracks[t].instrument.value = line.note_columns[nc].instrument_value+1

                 -- get real instrNumber from instrument_values (verschiebt sich wenn Multisamples in der Liste aufgefuehrt sind)
                 -- print ('instr in pattern='..line.note_columns[nc].instrument_value+1)
                 local index_in_iv = 1
                 for ins=1, #instrument_values do
                    local instr_in_iv = 0
                    local additional_sample_index = 0
                    local instrNewValue = instrument_values[ins]
                    local wert = string.find(instrNewValue, ',')
                    if (wert ~= nil) then -- multisample gefunden
                      instr_in_iv = 0 + string.sub(instrNewValue, 1, wert-1)
                      -- compare base note of sample with note in track. if matches add sample index to get correct sample# of multisample
                      for s=1, #renoise.song().instruments[instr_in_iv].samples do
                        if(renoise.song().instruments[instr_in_iv].samples[s].base_note==line.note_columns[nc].note_value) then
                          additional_sample_index=s-1
                          break
                        end
                      end
                    else
                      instr_in_iv = 0 + instrument_values[ins]                 
                    end
                    
                    if (line.note_columns[nc].instrument_value+1==instr_in_iv) then
                      index_in_iv = ins + additional_sample_index
                      break
                    end
                 end
                 tracks[t].instrument.value = index_in_iv

                 -- loading notes
                 tracks[t].note1.value = line.note_columns[nc].note_value+1
                 if not line.note_columns[nc+1].is_empty then
                   tracks[t].note2.value = line.note_columns[nc+1].note_value+1
                 end
                 if not line.note_columns[nc+2].is_empty then
                   tracks[t].note3.value = line.note_columns[nc+2].note_value+1
                 end
                                         
                 -- loading velocity, delay. length won't be loaded
                 tracks[t].velocity.value = line.note_columns[nc].volume_value
                 tracks[t].delay.value = line.note_columns[nc].delay_value              

                 has_loaded_parameters = true
               end

               -- setup the lights of each track
               tracks[t].lights[lpos].value=true
               break;
             end
             
           end  -- for lpos          
         
       end -- if line.is_empty
       else
          for lpos=1,#tracks[t].light_positions do
            --print ('amount light positions = ' .. #tracks[t].light_positions)          
            if tracks[t].light_positions[lpos].value==pos.line then
              --print ('Track t= ' .. t .. ' position=' .. pos.line .. ' tracklight pos=' .. tracks[t].light_positions[lpos].value .. ' lpos='..lpos)
              tracks[t].lights[lpos].value=false
              break;
            end
          end
       end

    end
  
  end

  set_notes = true;
end

-- create all tracks and save them into a table
tracks = { }

function create_tracks()
  tracks = { }
  for j=1, number_of_tracks do
   
    local instrnr = j
    if instrnr>=#renoise.song().instruments then instrnr = 1 end
    print ('Create Track Nr ' .. j .. ': Instrnr=' .. instrnr .. ' #instr= ' .. #renoise.song().instruments)
        
    -- the track document. It holds all necessary data for one track of the step sequencer (steps, velocity, etc.)
    local track = renoise.Document.create("StepSequencerTrack") {
      
      track_number = j;
      track_color =  {0,0,0};
      
      note1 =49;
      note2 =122;
      note3 =122;
      notelength = 0; -- when note OFF is shown. < editstep
      instrument = instrnr; -- instrument number
      
      lights = {false};
      light_positions = {1};
      velocity = 0x7F;
      delay = 0x00;
    }
     
    table.insert(tracks, j, track);
    --print('tracknumber val= ' .. tracks[j].track_number.value)
  end
end


-- init all lights and positions for tracks
function init_track_lights()
  print ('init track_lights()')
  for i=1,#tracks do
    local pos = 1
    for j=1,number_of_steps-1 do
      tracks[i].lights:insert(false);
      pos = pos + editstep
      tracks[i].light_positions:insert(pos);  
    end
  end
end

-- resets the position of the lights, regarding editpage
function reload_track_lights_positions()
  
  local pattern_end = editpage * number_of_steps * editstep
  local pattern_start = pattern_end - (number_of_steps * editstep)
  lauflicht_start = pattern_start+1;
  lauflicht_end = pattern_end;
  
  for i=1,#tracks do
    local pos = pattern_start + 1 -- because first position of pattern is 1, not 0
    for j=1,#tracks[i].light_positions do
      tracks[i].light_positions[j].value = pos
      --print('insert into lights positon ' .. j ..' pos:'.. pos)
      pos = pos + editstep
    end
    print ('track nr '..i.. ' -> Number of reloaded light positions:' .. #tracks[i].light_positions)
  end
end

-- resets the editstep for each track, if editstep was changed
function reset_Editstep() 
  for i=1,#tracks do  
    local pos = 1
    for j=1,number_of_steps do
      --tracks[i].lights:insert(false);
      tracks[i].light_positions[j].value = pos;
      pos = pos + editstep
    end
  end
end

-- resets the note length of each track, if editstep was changed
function reset_Notelength()
  for i=1,#tracks do
    vb.views["note_length"..i].max = editstep-1
    if (tracks[i].notelength.value > editstep-1) then
      tracks[i].notelength.value = 0
    end
  end
end


-- resets the editpage value
function reset_Editpage()
  local maxPage= (renoise.song().patterns[renoise.song().selected_pattern_index].number_of_lines/editstep)/number_of_steps
  if (maxPage<1) then maxPage=1 end
  vb.views["gui_editpage"].max = maxPage
  vb.views["gui_editpage"].value = 1
end

-- checks if position if valid. if not, shows error message
function is_position_valid(pos, showDialog)

  local patternLines=renoise.song().patterns[renoise.song().selected_pattern_index].number_of_lines

  if (pos>patternLines) then 
    if (showDialog) then
      showModalDialog("Error",  "Can not set note on position " .. pos-1 ..":\nThe position is not available in the pattern.\nMaximal position:"..patternLines-1)
    end
    return false
  end
  
  return true
  
end

-- set note-value
function set_note(set, pos, value1, value2, value3, track_nr, instr, volume, delay) 

  if (set==true) then
 
    -- get instr_nr from instrument_values ('1' oder bei multisamples '1,3')
    local instrNewValue = instrument_values[instr+1]
    local wert = string.find(instrNewValue, ',')
    if (wert ~= nil) then -- multisample gefunden
      instr = -1 + string.sub(instrNewValue, 1, wert-1)
    else
      instr = -1 + instrument_values[instr+1]
    end
        
    for columnnr=1,3 do
      local value = value1
      if (columnnr==2) then value = value2 end
      if (columnnr==3) then value = value3 end       
      --print('set note in column ' ..columnnr.. ' to ' .. value .. ' at pos ' .. pos)
      renoise.song().selected_pattern.tracks[track_nr].lines[pos].note_columns[columnnr].note_value = value; 
      if (value<120) then -- do not add volume etc if note is empty or OFF
        renoise.song().selected_pattern.tracks[track_nr].lines[pos].note_columns[columnnr].instrument_value = instr;  
        renoise.song().selected_pattern.tracks[track_nr].lines[pos].note_columns[columnnr].volume_value = volume;
        renoise.song().selected_pattern.tracks[track_nr].lines[pos].note_columns[columnnr].delay_value = delay
      end
    end
  else
    for columnnr=1,3 do
     -- print('delete notes at pos ' .. pos)      
      renoise.song().selected_pattern.tracks[track_nr].lines[pos].note_columns[columnnr].note_string = '---'; 
      renoise.song().selected_pattern.tracks[track_nr].lines[pos].note_columns[columnnr].instrument_value = 255; -- empty instr
      renoise.song().selected_pattern.tracks[track_nr].lines[pos].note_columns[columnnr].volume_string = '..'
      renoise.song().selected_pattern.tracks[track_nr].lines[pos].note_columns[columnnr].delay_string = '..'
    end
  end
end 

-- deletes a note off at given position in given track
function delete_noteOffs(pos, track_nr)
  
  for columnnr=1,3 do
    if (renoise.song().selected_pattern.tracks[track_nr].lines[pos].note_columns[columnnr].note_value==120) then
      renoise.song().selected_pattern.tracks[track_nr].lines[pos].note_columns[columnnr].note_string = '---';
    end
  end

end 

-- removes all notes of given track_nr regarding the editstep and page
function delete_allNotesOfTrack(track_nr) 
    
  for i=1,number_of_steps do
      tracks[track_nr].lights[i].value=false
  end
  
end


-- we do place our notifications (if needed now outside of the GUI code) for all tracks
function tracks_add_notifications()
  for i=1, #tracks do
    for j=1, #tracks[i].lights do 
        tracks[i].lights[j]:add_notifier(function()
        
        if not is_position_valid(tracks[i].light_positions[j].value, true) then
          return
        end
  
        if (set_notes==true) then
          set_note(tracks[i].lights[j].value, 
                   tracks[i].light_positions[j].value, 
                   tracks[i].note1.value-1,
                   tracks[i].note2.value-1,
                   tracks[i].note3.value-1,          
                   tracks[i].track_number.value,
                   tracks[i].instrument.value-1,
                   tracks[i].velocity.value,
                   tracks[i].delay.value)
          -- set note off note (note length)
          if (tracks[i].notelength.value>0) then
            local off1 = 120
            local off2 = 120
            local off3 = 120
            if (tracks[i].note1.value-1==121) then off1=121 end -- set only note off if notevalue was set.
            if (tracks[i].note2.value-1==121) then off2=121 end
            if (tracks[i].note3.value-1==121) then off3=121 end
          set_note(tracks[i].lights[j].value, 
                   tracks[i].light_positions[j].value+tracks[i].notelength.value, 
                   off1,
                   off2,
                   off3,          
                   tracks[i].track_number.value,
                   tracks[i].instrument.value-1,
                   tracks[i].velocity.value,
                   tracks[i].delay.value)
           end
           if (tracks[i].lights[j].value==false) then
              -- delete note offs from pos+1 to pos+editstep-1
              for noffs=1,editstep-1 do
                  if is_position_valid(tracks[i].light_positions[j].value+noffs, false) then
                    delete_noteOffs(tracks[i].light_positions[j].value+noffs, tracks[i].track_number.value)
                  end
              end       
           end        
        end
        
        end
        )
    end 
  end
end 


-- define the midi-mapping functions for each midi-controlable gui-control
-- each step of each track can be mapped to a midi-event "switch" (toggle a button)
-- when a midi-button is pressed, the notified function of this step will be executed
-- the Editstep and Page controls can be assigned to a knob or slider (midi event "absolute value")
-- the functions will automatically calculate the correct ranges for the appropriate values
function add_midi_mappings()

  -- all steps of each track
  for i=1, #tracks do
    for j=1, #tracks[i].lights do 
      
      if (renoise.tool():has_midi_mapping(string.format("de.cie-online.StepSequencer:Track%02d:Step%02d",i,j))~=true) then
      
        renoise.tool():add_midi_mapping{
          name = string.format("de.cie-online.StepSequencer:Track%02d:Step%02d",i,j), 
          invoke = function(message)
              --print("de.cie-online.StepSequencer: >> got midi_mapping message : Track " .. i .. " Step:" ..j)

              if (message:is_switch()) then
                tracks[i].lights[j].value=message.boolean_value
              elseif (message:is_abs_value()) then             
                if (message.int_value<64) then
                  tracks[i].lights[j].value=false
                else
                  tracks[i].lights[j].value=true
                end
              end

          end
        }
      
      end
      
    end
  end
  
  -- Editstep
  if (renoise.tool():has_midi_mapping("de.cie-online.StepSequencer:Editstep")~=true) then

    renoise.tool():add_midi_mapping{
      name = "de.cie-online.StepSequencer:Editstep",
      invoke = function(message)
      
        if (message:is_abs_value()) then
          -- print(("de.cie-online.StepSequencer:Editstep  message.int_value: %d)"):format(message.int_value))
          -- split 127 value into 3 possible ranges for an editstep.
          if (message.int_value>=0 and message.int_value<42) then
            vb.views["gui_editstep"].value=1
          elseif (message.int_value>=42 and message.int_value<84) then
            vb.views["gui_editstep"].value=2
          else
            vb.views["gui_editstep"].value=3
          end
        end

      end
    }

  end
  
  -- Editpage
  if (renoise.tool():has_midi_mapping("de.cie-online.StepSequencer:Page")~=true) then

    renoise.tool():add_midi_mapping{
      name = "de.cie-online.StepSequencer:Page",
      invoke = function(message)
      
        if (message:is_abs_value()) then

          local maxPage = (renoise.song().patterns[renoise.song().selected_pattern_index].number_of_lines/editstep)/number_of_steps;
          if (maxPage<1) then maxPage=1 end;
    
          local range = 127/maxPage;
          local pagevalue = math.ceil(message.int_value/range)
          if (pagevalue==0) then pagevalue=1 end
                
          --print(("de.cie-online.StepSequencer:Page  message.int_value: %d)"):format(message.int_value))
          --print("ausgerechneter wert fuer page=" .. pagevalue)          
          vb.views["gui_editpage"].value=pagevalue

        end

      end
    }
    
  end
 
end


-- GUI 
midi_notes = {}
note_table = {"C-", "C#","D-","D#","E-","F-","F#","G-","G#","A-","A#","B-"}

function build_note_table()

  local current_midi_note = 0

  for octave = 0,9 do
  
      for notepos = 1,12 do
        current_midi_note = current_midi_note +1
        midi_notes[current_midi_note] = note_table[notepos]..tostring(octave)
        
        if current_midi_note == 128 then
          break
        end
      end
      
      if current_midi_note == 128 then
        break
      end
  end
  midi_notes[current_midi_note+1] = 'Off' -- add note off
  midi_notes[current_midi_note+2] = '---' -- add empty note

end


-- get track names
track_names = {}
function build_track_names_table()
  track_names = {}
  for i=1,#renoise.song().tracks do 
    local track = renoise.song().tracks[i];
    if track.type==renoise.Track.TRACK_TYPE_SEQUENCER then
      --print('track#'..i.. ' name=' .. track.name .. ' color='..track.color[1])
      table.insert(track_names, renoise.song().tracks[i].name)
    end
  
  end
end

-- add notifier to each track name and update the Track_SelectBox if the name of a track changes
function track_names_add_notifier()

  for i=1,#renoise.song().tracks do 
    local track = renoise.song().tracks[i];
    if track.type==renoise.Track.TRACK_TYPE_SEQUENCER then
      track.name_observable:add_notifier(function()
        build_track_names_table()
        for j=1, #tracks do
            vb.views["track_number"..j].items = track_names 
        end
      end
      )
    end
  end

end

-- get instrument names
instrument_names = {}
instrument_values = {}
function build_instrument_names_table()
  instrument_names = {}
  instrument_values = {}
  print("build instr table")
  for i=1,#renoise.song().instruments do 
      --table.insert(instrument_names, renoise.song().instruments[i].name)
      
      --print('instr # ' .. i .. ' ' .. renoise.song().instruments[i].name)
      
      if (#renoise.song().instruments[i].samples>2) then
        --print ('Instrument is a multisample! Samples:')
        for s=1, #renoise.song().instruments[i].samples do
          --print('sample='..s.. ':' .. renoise.song().instruments[i].samples[s].name .. 
          --' basenote=' .. renoise.song().instruments[i].samples[s].base_note)
          table.insert(instrument_names, renoise.song().instruments[i].name..':'..renoise.song().instruments[i].samples[s].name)
          table.insert(instrument_values, i..','..s)
        end
      else
        table.insert(instrument_names, renoise.song().instruments[i].name)
        table.insert(instrument_values, ''..i)
      end      
  end
end

-- add notifier to each instr name and update the Instrument_SelectBox if the name of a instr changes
function instrument_names_add_notifier()
  for i=1,#renoise.song().instruments do 
      renoise.song().instruments[i].name_observable:add_notifier(function()
        build_instrument_names_table()
        for j=1,#tracks do
            vb.views["instrument_number"..j].items = instrument_names 
        end
      end
      )
  end
end

function instrument_add_notifier()
  renoise.song().instruments_observable:add_notifier(function()
    instrument_names_add_notifier()
    build_instrument_names_table()
    for j=1,#tracks do
      vb.views["instrument_number"..j].items = instrument_names 
    end
  
  end)
end



-- TODO: if patternsize mod 16 != 0 or < 16 warning message
-- TODO: on current pattern change recall this method
function patternSize_add_notifier()
  renoise.song().patterns[renoise.song().selected_pattern_index].number_of_lines_observable:add_notifier(function()
    print('pattern size changed to ' .. renoise.song().patterns[renoise.song().selected_pattern_index].number_of_lines)
    end)
end

-- if pattern index changes
function patternIndex_add_notifier()
 renoise.song().selected_pattern_index_observable:add_notifier(function()
      --print ('changed pattern idx to = ' .. renoise.song().selected_pattern_index)
      load_pattern_into_sequencer() -- reload pattern data into sequencer (lights and note values)
      end
)

end

function trackAmountChange_add_notifier()

  renoise.song().tracks_observable:add_notifier(function()

  -- alle tracklights rows von maingui removen
  -- for all tracks t do:
  -- alle vb.views['track_number'..i] und ALLE anderen GUI-parts mit id == nil setzen
  -- dann alles wieder pro track neu erzeugen.
    
  print ('vor removing track rows ' .. number_of_tracks)  
  local number_of_tracks_new=0;
  for i=1,#renoise.song().tracks do 
    local track = renoise.song().tracks[i];
    if track.type==renoise.Track.TRACK_TYPE_SEQUENCER then
      number_of_tracks_new=number_of_tracks_new+1
    end
  end
  
  if (#tracks~=number_of_tracks_new) then
   
      for t=1, number_of_tracks do
        vb.views['track_rows']:remove_child(vb.views['column'..t])
        vb.views['column'..t]=nil
      end
      
      for f=1, number_of_tracks do
        vb.views['track_number'..f]=nil
        vb.views['instrument_number'..f]=nil
        vb.views['note_length'..f]=nil
        vb.views['delete_all'..f]=nil
      end
      
      for b=1,number_of_steps do
        vb.views['lauflicht'..b]=nil
      end
         
      vb.views['maingui']:remove_child(vb.views['text_row'])
      vb.views['text_row']=nil 
      vb.views['maingui']:remove_child(vb.views['track_rows'])
      vb.views['track_rows']=nil 
      vb.views['maingui']:remove_child(vb.views['prefs_row'])
      vb.views['prefs_row']=nil
      vb.views['gui_editstep']=nil
      vb.views['gui_editpage']=nil
      
      -- remove midi_mappings
      for c=1, number_of_tracks do
        for d=1,number_of_steps do
          renoise.tool():remove_midi_mapping(string.format("de.cie-online.StepSequencer:Track%02d:Step%02d",c,d))
        end
      end

         
      print ('changed amount of tracks to= ' .. #renoise.song().tracks)

      load_tracks()
      create_tracks()
      init_track_lights()
      tracks_add_notifications()
      build_track_names_table()
      
      create_gui_contents()
      
      reload_track_lights_positions()
      load_pattern_into_sequencer()
      
      add_midi_mappings()

    end

  end
  )

end


-- if a new instrument was chosen from drop-down box, parse pattern and replace all notes of the track
-- with the new selected instrument and note 
function change_instr_of_track(t) 

  local pattern_iter = renoise.song().pattern_iterator
  local pattern_index =  renoise.song().selected_pattern_index
  
   print ('Load Notes into Sequencer of Track ' ..t .. ' number of tracks=' ..number_of_tracks)

   for pos,line in pattern_iter:lines_in_pattern_track(pattern_index,t) do


     local nc=1
     if not line.is_empty and line.note_columns[nc].note_value~=121 then -- do not parse lines with no note value
     
        if not line.note_columns[nc].is_empty then 
            
          -- set instrument to new selected instrument and note value
          -- get instr_nr from instrument_values ('1' oder bei multisamples '1,3')
          
          local instr = 1
          local instrNewValue = instrument_values[tracks[t].instrument.value]
          local wert = string.find(instrNewValue, ',')
          if (wert ~= nil) then -- multisample gefunden
            instr = -1 + string.sub(instrNewValue, 1, wert-1)
          else
            instr = -1 + instrument_values[tracks[t].instrument.value]
          end
          line.note_columns[nc].instrument_value = instr
          
          print ('instrval=' .. tracks[t].instrument.value .. ' new val=' .. instr)
            
          -- set notes to new notes (if a multisample was chosen from instrument list)
          line.note_columns[nc].note_value = tracks[t].note1.value-1;
          if not line.note_columns[nc+1].is_empty then
            line.note_columns[nc+1].note_value = tracks[t].note2.value-1
          end
          if not line.note_columns[nc+2].is_empty then
            line.note_columns[nc+2].note_value = tracks[t].note3.value-1
          end                   
                   
        end
     end
     
   end
end


-- shows a modal dialog with given title and text
-- in first line for error messages
function showModalDialog(title, text)

  local vb = renoise.ViewBuilder()
  local dialog_title = title
  local dialog_buttons = {"OK"};

  local DEFAULT_MARGIN = renoise.ViewBuilder.DEFAULT_CONTROL_MARGIN

  local dialog_content = vb:column {
    margin = DEFAULT_MARGIN,
    vb:column {
      style = "group",
      margin = DEFAULT_MARGIN,
      vb:text {
        text = text
      },
    }
  }

  renoise.app():show_custom_prompt(dialog_title, dialog_content, dialog_buttons)
end




-- main function, will create the gui of the step sequencer  
function create_gui()

  load_tracks();
  print ('create gui: tracks amount=' .. #tracks .. ' numberOftracks='..number_of_tracks);
  if (#tracks~=number_of_tracks) then -- only init sequencer on first start. Keep all values when close and restart.
    -- print ('init-track-lights=' .. #tracks);
    create_tracks()
    init_track_lights()
    tracks_add_notifications()
  end
  
    trackAmountChange_add_notifier()
    build_note_table()
    build_track_names_table()
    build_instrument_names_table()
  
    track_names_add_notifier()
    instrument_names_add_notifier()

    instrument_add_notifier() -- if new samples are loaded
  
    reload_track_lights_positions()
    load_pattern_into_sequencer()
    patternSize_add_notifier()
    patternIndex_add_notifier()
       
end


function create_tracks_for_gui()
   -- add for every track a lights row and bind it to the track document 
  local lights_rows = { } 
  for r=1, number_of_tracks do
    table.insert(lights_rows, vb:row{})
  end
  
  for i=1, number_of_tracks do

    --print('track_tracknumer= ' .. tracks[i].track_number.value)
    -- track number/name
    local track_number = vb:popup {
      id = 'track_number'..i,
      bind = tracks[i].track_number,
      value = tracks[i].track_number.value, --TODO: anderen wert festlegen, z.b. '1,3'
      width = 80,
      items = track_names
    }
    lights_rows[i]:add_child(track_number)

    -- instrument number/name
    local instrument_number = vb:popup {
      id = 'instrument_number'..i,
      bind = tracks[i].instrument,
      value = tracks[i].instrument.value,
      width = 120,
      items = instrument_names,
      notifier=function()
        local newValue = vb.views["instrument_number"..i].value
        --print('selected value = ' .. vb.views["instrument_number"..i].value .. 'instr value=' .. instrument_values[newValue])
        local instrNewValue = instrument_values[newValue]
        local instrNr = instrNewValue
        local sampleNr = 1
        --tracks[i].note1.value = 49 -- default note when changing instrument
        
        -- instrument_values '1' oder bei multisamples '1,3' zerlegen
        local wert = string.find(instrNewValue, ',')
        if (wert ~= nil) then -- multisample gefunden
          instrNr = string.sub(instrNewValue, 1, wert-1)
          sampleNr = string.sub(instrNewValue, wert+1)
          local basenote=renoise.song().instruments[0+instrNr].samples[0+sampleNr].base_note
          --print('base note='.. basenote.. ' ' .. renoise.song().instruments[0+instrNr].samples[0+sampleNr].name)
          tracks[i].note1.value = basenote+1 -- set note to sample basenote
        end
        change_instr_of_track(i)
      end
    }
    lights_rows[i]:add_child(instrument_number)    
        
    -- steps: bind every step to a notifier (calls a function) and to a midi-mapping (calls the notifier function)
    for j=1, #tracks[i].lights do
      local cb = vb:checkbox {
        bind = tracks[i].lights[j], 
        midi_mapping = string.format("de.cie-online.StepSequencer:Track%02d:Step%02d",i,j)
      }
      lights_rows[i]:add_child(cb)
    end
    

    -- velocity
    local velocitybox = vb:valuebox{
      width = 50,
      bind = tracks[i].velocity, -- bind
      min = 0, 
      max = 0x7f,
      tostring = function(value) 
        return ("%.2X"):format(value)
      end,
      tonumber = function(str) 
        return tonumber(str, 0x10)
      end      
    }
    lights_rows[i]:add_child(velocitybox)
    
    -- delay
    local delaybox = vb:valuebox{
      width = 50,
      bind = tracks[i].delay, -- bind
      min = 0, 
      max = 0xff,
      tostring = function(value) 
        return ("%.2X"):format(value)
      end,
      tonumber = function(str) 
        return tonumber(str, 0x10)
      end      
    }
    lights_rows[i]:add_child(delaybox)
     
    -- Note1
    local note_one = vb:popup {
      bind = tracks[i].note1,
      value = tracks[i].note1.value,
      width = 50,
      items = midi_notes
    }
    lights_rows[i]:add_child(note_one)
    
    -- Note2
    local note_two = vb:popup {
      bind = tracks[i].note2,
      value = tracks[i].note2.value,
      width = 50,
      items = midi_notes
    }
    lights_rows[i]:add_child(note_two)
    
    -- Note3
    local note_three = vb:popup {
      bind = tracks[i].note3,
      value = tracks[i].note3.value,
      width = 50,
      items = midi_notes
    }
    lights_rows[i]:add_child(note_three)        

    -- note length -> when setting note off after note(s) position
    local note_length = vb:valuebox {
      id = 'note_length'..i,
      bind = tracks[i].notelength,
      value = tracks[i].notelength.value,
      min = 0,
      max = editstep-1,
      width = 50,
    }
    lights_rows[i]:add_child(note_length)
    
    local space = vb:space {
      width = 10
    }
    lights_rows[i]:add_child(space)
  
    local delete_all = vb:button {
      id = 'delete_all'..i,
      color = {0x7e,0x3b,0x3b},
      notifier=function()
      delete_allNotesOfTrack(i)
    end    
    }
    lights_rows[i]:add_child(delete_all)
             
  end  
  
    -- gui block which holds all track rows
  local track_rows = vb:column{ id='track_rows'}
  
  -- insert for every track a lights_row into track_rows
  for t=1, number_of_tracks do
    local column = vb:column { id = 'column'..t}
    column:add_child(lights_rows[t])
    vb.views['track_rows']:add_child(column)
  end
  -- now insert track_rows into the main_gui
  vb.views['maingui']:add_child(track_rows)
  
  
end



function create_gui_dialog() 

  
  local DIALOG_MARGIN = renoise.ViewBuilder.DEFAULT_DIALOG_MARGIN
  local CONTENT_SPACING = renoise.ViewBuilder.DEFAULT_CONTROL_SPACING
  
  renoise.app():show_custom_dialog("Lauflicht Step Sequencer "..version, 
    vb:column {
      margin = DIALOG_MARGIN,
      uniform = true,
      id='maingui',
 
      vb:column {
        spacing = CONTENT_SPACING,
        --text_row
      }      
    }      
  )
  
end

function create_gui_contents()

 -- text labels row
  local text_row = vb:row{ id='text_row'}
  local label_track = vb:text {
    text="Track",
    align = "center",
    width=80
  }
  local label_instrument = vb:text {
    text="Instr",
    align = "center",
    width=120
  }        
  local space = vb:space {
      height = 20,
      --width = 290
  }
  local label_velocity = vb:text {
    text="Velocity",
    align = "center",
    width=50
  }
  local label_delay = vb:text {
    text="Delay",
    align = "center",
    width=50
  }
  local label_note1 = vb:text {
    text="Note 1",
    width=50
  }  
  local label_note2 = vb:text {
    text="Note 2",
    width=50
  }  
  local label_note3 = vb:text {
    text="Note 3",
    width=50
  } 
  local label_notelength = vb:text {
    text="Length",
    width=50
  }         
  local label_deleteRestore = vb:text {
    text="Delete",
    width=50
  }    
  
  text_row:add_child(label_track)
  text_row:add_child(label_instrument) 
  
  -- the lauflicht/moving light sequencer, consisting of #number_of_steps buttons.
  -- add to every button an id to identify button anywhere in the program with vb.views[id] e.g. vb.views["lauflicht1"].color={0x00, 0x30, 0x10}
  for b=1,number_of_steps do
    local but = vb:button { id="lauflicht"..b}
    text_row:add_child(but);
  end
 
  --text_row:add_child(space)
  text_row:add_child(label_velocity)
  text_row:add_child(label_delay)
  text_row:add_child(label_note1)
  text_row:add_child(label_note2)
  text_row:add_child(label_note3)  
  text_row:add_child(label_notelength)
  text_row:add_child(label_deleteRestore)
  text_row:add_child(space)
  
  -- preferences row
  local prefs_row = vb:row{ id='prefs_row'}
  local label_editstep = vb:text {
    text="Editstep",
    align = "center",
    width=80
  }
  local gui_editstep = vb:popup {
    id = 'gui_editstep',
    value = editstepValue,
    width = 80,
    items = {'1','2','4'}, -- editstep 8 buggy because of 16*8 > 64 steps
    midi_mapping = "de.cie-online.StepSequencer:Editstep",
    notifier=function()
      editstep = vb.views["gui_editstep"].value
      editstepValue = vb.views["gui_editstep"].value
      if (vb.views["gui_editstep"].value == 3) then editstep=4 end -- convert sent values 3 and 4 to right editstep values
      if (vb.views["gui_editstep"].value == 4) then editstep=8 end
      -- print('editstep=' ..  vb.views["gui_editstep"].value .. ' ' .. editstep)
      reset_Editstep() -- set new editstep for each track
      reset_Notelength()-- set new maximal note length
      reset_Editpage() -- set new maximal page and reset page to 1
      reload_track_lights_positions()
      load_pattern_into_sequencer() -- reload pattern data into sequencer (lights and note values)
    end
  }  
  prefs_row:add_child(label_editstep)
  prefs_row:add_child(gui_editstep)

  local label_editpage = vb:text {
    text="Page",
    align = "center",
    width=80
  }
    
  local maxPage =  (renoise.song().patterns[renoise.song().selected_pattern_index].number_of_lines/editstep)/number_of_steps;
  if (maxPage<1) then maxPage=1 end;
  
  local gui_editpage = vb:valuebox {
    id = 'gui_editpage',
    value = editpage,
    min = 1,
    max = maxPage, --  (renoise.song().patterns[renoise.song().selected_pattern_index].number_of_lines/editstep)/number_of_steps,
    width = 50,
    midi_mapping = "de.cie-online.StepSequencer:Page",
    notifier=function()
      editpage = vb.views["gui_editpage"].value
      --print('editpage=' ..  vb.views["gui_editpage"].value)
      reload_track_lights_positions() -- reloads all lights positions
      load_pattern_into_sequencer() -- reload pattern data into sequencer (lights and note values)
    end    
  }
  prefs_row:add_child(label_editpage)
  prefs_row:add_child(gui_editpage) 

  --[[local label_credits = vb:text {
    text="by Cie",
    align = "right",
    width=520
  }
  prefs_row:add_child(label_credits) ]]
  
  vb.views['maingui']:add_child(text_row)
  create_tracks_for_gui()
  -- insert the preferences row 
  vb.views['maingui']:add_child(prefs_row)    
  
  print("Successfully build gui.")
  
end


