--[[----------------------------------------------------------------------------
de.cie-online.StepSequencer.xrnx/main.lua

LAUFLICHT STEP SEQUENCER PLUGIN FOR RENOISE
written and (c) by Thilo Geertzen alias Cie in 2010. http://www.cie-online.de

----------------------------------------------------------------------------]]--

require "gui"; 

--------------------------------------------------------------------------------
-- preferences

-- tools can have preferences, just like Renoise. To use them we first need 
-- to create a renoise.Document object which holds the options that we want to 
-- store/restore
local options = renoise.Document.create("StepSequencerOptions") {
  show_debug_prints = true
}

-- then we simply register this document as the main preferences for the tool:
renoise.tool().preferences = options

--------------------------------------------------------------------------------
-- menu entries

renoise.tool():add_menu_entry {
  name = "--- Main Menu:Tools:Step Sequencer",
  invoke = function() 
     vb = renoise.ViewBuilder()
     create_gui_dialog()
     create_gui()
     create_gui_contents()
     add_midi_mappings()
  end
}

--------------------------------------------------------------------------------
-- notifications

-- invoked, as soon as the application became the foreground window,
-- for example when you alt-tab'ed to it, or switched with the mouse
-- from another app to Renoise
renoise.tool().app_became_active_observable:add_notifier(function()
 handle_app_became_active_notification()
end)
  
-- invoked, as soon as the application looses focus, another app
-- became the foreground window
renoise.tool().app_resigned_active_observable:add_notifier(function()
  handle_app_resigned_active_notification()
end)
  
-- invoked periodically in the background, more often when the work load
-- is low. less often when Renoises work load is high.
-- The exact interval is not defined and can not be relied on, but will be
-- around 10 times per sec.
-- You can do stuff in the background without blocking the application here.
-- Be gentle and don't do CPU heavy stuff in your notifier!
renoise.tool().app_idle_observable:add_notifier(function()
  handle_app_idle_notification()
end)
  
-- invoked each time a new document (song) was created or loaded
renoise.tool().app_new_document_observable:add_notifier(function()
  handle_app_new_document_notification()
end)


--------------------------------------------------------------------------------
-- debug hook

_AUTO_RELOAD_DEBUG = function()
  handle_auto_reload_debug_notification()
end

-------------------------------------------------------------------------------

if (options.show_debug_prints.value) then
  print("de.cie-online.StepSequencer: script was loaded...")
end


-------------------------------------------------------------------------------

-- implementation if the nofification callbacks, as attached above...

-- handle_app_became_active_notification
function handle_app_became_active_notification()
  if (options.show_debug_prints.value) then
    print("de.cie-online.StepSequencer: >> app_became_active notification")
  end
end


-- handle_app_resigned_active_notification
function handle_app_resigned_active_notification()
  if (options.show_debug_prints.value) then
    print("de.cie-online.StepSequencer: << app_resigned_active notification")
  end
end


-- handle_app_idle_notification
local last_idle_time = os.clock()
local light_number = 1
function handle_app_idle_notification()

  if (renoise.song().transport.playing and vb.views.maingui~=nil) then
    if os.clock() - last_idle_time >=0.01 then      

      local position = renoise.song().transport.playback_pos.line
      last_idle_time = os.clock();
      local lauflicht_length = number_of_steps;
      local lauflicht_step = lauflicht_start;
      if(renoise.song().patterns[renoise.song().selected_pattern_index].number_of_lines<lauflicht_length) then
        lauflicht_length=renoise.song().patterns[renoise.song().selected_pattern_index].number_of_lines
      end
      
      for i=1,lauflicht_length do
        vb.views["lauflicht"..lauflicht_length].color = {0x41,0x4c,0x43}
        if (position>=lauflicht_step and position<=lauflicht_end) then
           
           local current_light = i;
           local prev_light = i-1;
           if (i==1) then
             prev_light=lauflicht_length;
           end
                    
           vb.views["lauflicht"..prev_light].color = {0x41,0x4c,0x43}
           vb.views["lauflicht"..current_light].color = {0x11,0xcc,0x33}
        else 
           vb.views["lauflicht"..i].color = {0x41,0x4c,0x43}
        end
        
        lauflicht_step = lauflicht_step + editstep;
      end

    end
    
  end
  
end


-- handle_app_new_document_notification
function handle_app_new_document_notification()
  if (options.show_debug_prints.value) then
    print("de.cie-online.StepSequencer: !! app_new_document notification")
  end
end


-- handle_auto_reload_debug_notification
function handle_auto_reload_debug_notification()
  if (options.show_debug_prints.value) then
    print("de.cie-online.StepSequencer: ** auto_reload_debug notification")
  end
end


--[[----------------------------------------------------------------------------
----------------------------------------------------------------------------]]--
